package betir;

import java.util.*;
import java.io.*;
import java.awt.geom.Point2D;

/**
*	This is the base class that runs each player and decides which behavior will fire.
*/
public class Player implements Runnable
{
//****************************************************************************************************************** member variables

	/** The robocup environment*/
	RobocupEnvironment m_Env;
	/** The number of the player*/
	int m_nPlayerNumber;
	/** The class that sends the messages to the soccer server*/
	DatagramWrapper m_dgWrapper;
	/** The player thread*/
	private Thread m_Thread;
	/** Vector of behaviors that the plater has*/
	private Vector m_vBehaviors;
	/** Vector of behaviors that are able to fire*/
	private Vector m_vMatches;
	/** Vector of behaviors that are able to fire and are not inhibited*/
	private Vector m_vUninhibited;

	private int m_nYellRemove = 0;
	private int m_nLastTime;
	private long m_lTime;

	/** The start position of the player*/
	protected TwoDVec m_vecStart;
	/** The zone position of the player*/
	protected TwoDVec m_vecZone;

//****************************************************************************************************************** constructors

	/**
	*	The Constructor initializes the zone and playernumber and starts the player thread
	*
	*	@params theEnv 		an already created robocup environment
	*	@params dg			the already created class that sends messages to the soccer server
	*	@params nPlayerNum	the number of the player
	*	@params nZone 		the zone that the player will occupy
	*/
	public Player(RobocupEnvironment theEnv, DatagramWrapper dg, int nPlayerNum, int nZone)
	{
		m_Env = theEnv;
		m_nPlayerNumber = nPlayerNum;
		m_Env.m_nZone = nZone;
		m_dgWrapper = dg;

		m_vBehaviors = new Vector();
		m_vMatches = new Vector();
		m_vUninhibited = new Vector();

		m_Env.m_Player = this;

		switch(nZone)
		{
		case 0: //top
			m_vecStart = new TwoDVec(-5, 25, 0, 0);
			m_vecZone = new TwoDVec(5, 25, 0, 0);
			break;
		case 1: //bottom
			m_vecStart = new TwoDVec(-5, -25, 0, 0);
			m_vecZone = new TwoDVec(5, -25, 0, 0);
			break;
		case 2: //top goal
			m_vecStart = new TwoDVec(-38, 15, 0, 0);
			m_vecZone = new TwoDVec(-38, 15, 0, 0);
			break;
		case 3: //bottom goal
			m_vecStart = new TwoDVec(-38, -15, 0, 0);
			m_vecZone = new TwoDVec(-38, -15, 0, 0);
			break;
		case 4: //back center
			m_vecStart = new TwoDVec(-30, 0, 0, 0);
			m_vecZone = new TwoDVec(-30, 0, 0, 0);
			break;
		case 5: //center
			m_vecStart = new TwoDVec(0, 0, 0, 0);
			m_vecZone = new TwoDVec(0, 0, 0, 0);
			break;
		case 6: // top attacker
			m_vecStart = new TwoDVec(-1, 10, 0, 0);
			m_vecZone = new TwoDVec(40, 6, 0, 0);
			break;
		case 7: //bottom attacker
			m_vecStart = new TwoDVec(-1, -10, 0, 0);
			m_vecZone = new TwoDVec(40, -6, 0, 0);
			break;
		case 8: //center top attacker
			m_vecStart = new TwoDVec(-8, -4, 0, 0);
			m_vecZone = new TwoDVec(-8, -4, 0, 0);
			break;
		case 9: //center bottom attacker
			m_vecStart = new TwoDVec(-8, 4, 0, 0);
			m_vecZone = new TwoDVec(-8, 4, 0, 0);
			break;

		case 10: //goalie
			m_vecStart = new TwoDVec(-51, 0, 0, 0);
			m_vecZone = new TwoDVec(-51, 0, 0, 0);
			break;
		default:
			m_vecStart = new TwoDVec(-15, (int)(Math.random() * 16) - 8, 0, 0);
			m_vecZone = new TwoDVec(-15, (int)(Math.random() * 16) - 8, 0, 0);
			break;
		}

		m_Thread = new Thread(this, "AM-" + nPlayerNum);
		m_Thread.start();
	}

//****************************************************************************************************************** behavior stuff

	/**
	*	Adds a behavior to the player
	*
	*	@param rb	the behavior to be added
	*/
	public void addBehavior(RobocupBehavior rb) 	{ m_vBehaviors.add(rb); }

	/**
	*	Removes a behavior from the player
	*
	*	@param rb	the behavior to be removed
	*/
    public void removeBehavior(RobocupBehavior rb) 	{ m_vBehaviors.remove(rb); }

//****************************************************************************************************************** run

	/**
	*	Fires behaviors every 100 ms
	*/
	public void run()
	{
		m_nLastTime = m_Env.m_nCurrentTime;
		m_lTime = System.currentTimeMillis();
		while(true)
		{
			try
			{
				//parse all input info into the environment
				m_Env.parse(m_dgWrapper.receive(10));
			}
			catch (EOFException e)
            {
                //soccerserver died, quit
                System.err.println("Soccerserver died. Agent " + m_nPlayerNumber + " quiting.");
                break;
            }
            catch (InterruptedIOException e)
            {
                //receive timed out, keep going
            };

			if((System.currentTimeMillis() - m_lTime) > 100) //only fire a behavior every 100ms
			{
				m_lTime = System.currentTimeMillis();
				m_nLastTime = m_Env.m_nCurrentTime;

				//check which can fire
				m_vMatches.clear();
				for(int i = 0; i < m_vBehaviors.size(); ++i)
				{
					RobocupBehavior rb = (RobocupBehavior)m_vBehaviors.get(i);
					if(rb.canFire())
						m_vMatches.addElement(rb);
				}

				//check inhibitions
				m_vUninhibited.clear();
				for(int i = 0; i < m_vMatches.size(); ++i)
				{
					RobocupBehavior rb = (RobocupBehavior)m_vMatches.get(i);
					boolean bInhibited = false;
					for(int j = 0; j < m_vMatches.size(); ++j)
					{
						RobocupBehavior rb2 = (RobocupBehavior)m_vMatches.get(j);
						if(rb.equals(rb2))
							continue;
						if(rb2.inhibits(rb))
						{
							bInhibited = true;
							break;
						}
					}
					if(!bInhibited)
						m_vUninhibited.addElement(rb);
				}

				for(int i = 0; i < m_vUninhibited.size(); ++i)
				{
					RobocupBehavior rb = (RobocupBehavior)(m_vUninhibited.elementAt(i));
					if(rb.fire())
						removeBehavior(rb);
				}

				//update the ball, it gets overwritten if new data comes in
				if(m_Env.m_objBall != null)
				{
					m_Env.m_objBall.setAngle(m_Env.m_objBall.getAngle() + m_Env.m_objBall.getAngleChng());
					m_Env.m_objBall.setDistance(m_Env.m_objBall.getDistance() + m_Env.m_objBall.getDistChng());
				}

				//System.out.println("Behavior Firing is taking " + (System.currentTimeMillis() - m_lTime) + "ms");
				//m_Env.debug();
			}
		}
	}
}
