package betir;

/**
*	This is the main behavior class.
*/
public abstract class RobocupBehavior
{
//********************************************************************************************* basic stuff
	/** the robocup environment*/
	RobocupEnvironment m_Env;
	/** the class that sends messages to the soccerserver*/
	DatagramWrapper m_dgWrapper;

	public RobocupBehavior(RobocupEnvironment env, DatagramWrapper dg)
	{
		m_Env = env;
		m_dgWrapper = dg;
	}

	/** Determines if a behavior can fire or not*/
	public abstract boolean canFire();
	/**
	*	Determines if this behavior inhibits another behavior
	*
	*	@param rb the behavior to be checked
	*/
	public abstract boolean inhibits(RobocupBehavior rb);
	/**	Execute the behavior*/
	public abstract boolean fire();


//********************************************************************************************* server commands

	/**
	*	Sends a move message to the soccer server
	*
	*	@param 	nX	the x-corrdinate
	*	@param 	nY	the y-corrdinate
	*/
	public void move(int nX, int nY)
	{
		m_dgWrapper.send("(move " + Integer.toString(nX) + " " + Integer.toString(nY) + ")");
		m_Env.m_aobjFriends.clear();
		m_Env.m_bAbsAngle = false;
	}

	/**
	*	Sends a dash command to the server
	*
	*	@param	fPower	the power to dash
	*/
	public void dash(float fPower)
	{
		m_dgWrapper.send("(dash " + Float.toString(fPower) + ")");
		m_Env.m_bAbsAngle = false;
	}

	/**
	*	Sends a kick message to the soccer server
	*
	*	@param fPower	the power to kick
	*	@param nAngle	the angle to kick
	*/
	public void kick(float fPower, int nAngle)
	{
		m_dgWrapper.send("(kick " + Float.toString(fPower) + " " + Integer.toString(nAngle) + ")");
	}

	/**
	*	Sends a turn command to the server
	*
	*	@param nAngle	the angle to turn your body
	*/
	public void turn(int nAngle)
	{
		m_dgWrapper.send("(turn " + Integer.toString(nAngle) + ")");
		m_Env.m_vecPosition.rotPolar(-nAngle);
		m_Env.m_aobjFriends.clear();
		m_Env.m_objGoal = null;
		if(m_Env.m_objBall != null)
			m_Env.m_objBall.setAngle(m_Env.m_objBall.getAngle() - nAngle);
		m_Env.m_bAbsAngle = false;
	}

	/**
	*	Sends a turn_neck command to the server
	*
	*	@param	nAngle	the angle to turn
	*/
	public void turnHead(int nAngle)
	{
		m_dgWrapper.send("(turn_neck " + Integer.toString(nAngle) + ")");
		m_Env.m_objSelf.setHeadAngle(m_Env.m_objSelf.getHeadAngle() + nAngle);
		m_Env.m_vecPosition.rotPolar(-nAngle);
		m_Env.m_aobjFriends.clear();
		m_Env.m_objGoal = null;
		if(m_Env.m_objBall != null)
			m_Env.m_objBall.setAngle(m_Env.m_objBall.getAngle() - nAngle);
		m_Env.m_bAbsAngle = false;
	}

	/**
	*	Sends a say command to the server
	*
	*	@param str	the message to say
	*/
	public void say(String str)
	{
		if(str == null)
			return;
		if(str.charAt(0) == '"')
			str = str.substring(1, str.length() - 2);
		//System.out.println(str);
		m_dgWrapper.send("(say " + str + ")");
	}

	/**
	*	Sends a catch command to the server
	*
	*	@param	direction 	the direction to catch the ball
	*/
	public void catchBall(int direction)
	{
		m_dgWrapper.send("(catch " + direction + ")");
	}


	public void changeView(String strWidth, String strQuality)
	{
		if(!m_Env.m_objSelf.getViewWidth().equals(strWidth) || !m_Env.m_objSelf.getViewQuality().equals(strQuality))
			m_dgWrapper.send("(change_view " + strWidth + " " + strQuality + ")");
	}

//********************************************************************************************* utility finctions

	/**
	*	@return int		a players distance from his zone
	*/
	public int getZoneDist()
	{
		int nX = m_Env.m_Player.m_vecZone.getX();
		int nY = m_Env.m_Player.m_vecZone.getY();
		if(m_Env.m_objSelf.getSide() == 'r')
		{
			nX += m_Env.m_vecPosition.getX();
			nY -= m_Env.m_vecPosition.getY();
		}
		else
		{
			nX -= m_Env.m_vecPosition.getX();
			nY -= m_Env.m_vecPosition.getY();
		}
		return QuickIntMath.Sqrt(nX * nX + nY * nY);
	}

	/**
	*	@param	nX	x-coordinate
	*	@param  nY  y-coordinate
	*	@return int	the angle from the player to the specified point
	*/
	public int angleToPoint(int nX, int nY)
	{
		if(m_Env.m_objSelf.getSide() == 'r')
			nX = -nX;
		nX = nX - m_Env.m_vecPosition.getX();
		nY = nY - m_Env.m_vecPosition.getY();

		double angle = 0;
		if(nX == 0)
			angle = (nY > 0) ? 90 : -90;
		else if(nY == 0)
		  	angle = (nX > 0) ? 0 : 180;
		else
		{
      		angle = -Math.toDegrees(Math.atan((float)(nY) / (float)(nX)));
			if(nX < 0)
				angle -= 180;
		}
		angle += m_Env.m_vecPosition.getAngle();
		if(angle > 180) angle -= 360;
		if(angle < -180) angle += 360;
		return (int)(angle) + m_Env.m_objSelf.getHeadAngle();
    }

    /**
    *	@param  nAngle
    *	@return	int 	the angle from the player to a specified angle
    */
    public int angleToAngle(int nAngle)
    {
		int angle = m_Env.m_vecPosition.getAngle() - nAngle + m_Env.m_objSelf.getHeadAngle();
		if(angle > 180) angle -= 360;
		if(angle < -180) angle += 360;
		return angle;
	}

	/**
	*	@param	nNum	the number to check
	*	@param	nCenter	the center of the check
	*	@param	nRange 	the specified range
	*	@return boolean
	*/
	public static boolean withinRange(int nNum, int nCenter, int nRange)
	{
		if(nRange < 0)
			nRange = 0;
		return (Math.abs(nNum - nCenter) <= nRange);
	}

//********************************************************************************************* dash functions

	protected boolean m_bLastWasTurn = false;
	/**
	*	@param nX			the x-coordinate to dash to
	*	@param nY			the y-coordinate to dash to
	*	@param nMaxPower	the maximium power of the dash
	*/
    public void dashToPoint(int nX, int nY)
    {
		int fX = nX - m_Env.m_vecPosition.getX();
		int fY = nY - m_Env.m_vecPosition.getY();
		int nD = QuickIntMath.Sqrt(fX * fX + fY * fY);

		double dashPower = 5 *  Math.min(nD / m_Env.m_cfgData.dash_power_rate, m_Env.m_cfgData.maxpower);

		if (dashPower > m_Env.m_objSelf.getStamina() - m_Env.m_cfgData.recover_dec_thr * m_Env.m_cfgData.stamina_max - 100)
			dashPower = m_Env.m_objSelf.getStamina() - m_Env.m_cfgData.recover_dec_thr * m_Env.m_cfgData.stamina_max - 100;

		if (dashPower < 0)
      		dashPower = 0;

		if(m_bLastWasTurn || withinRange(angleToPoint(nX, nY), 0, nD / 10))
		{
			dash((float)dashPower);
			m_bLastWasTurn = false;
		}
		else
		{
			turn(angleToPoint(nX, nY));
			m_bLastWasTurn = true;
		}
	}

	/**
	*	Dashes to the ball.  This function uses prediction to attempt to optimize the interception of the ball.
	*/
	public void dashToBall()
	{
		if(m_Env.m_objBall == null)
			return;

		TwoDVec vec = findInterceptBall(30);

		if(Math.abs(m_Env.m_objBall.getAngleChng()) > 10)
			m_bLastWasTurn = false;

		dashToPoint(vec.getX(), vec.getY());
	}

	private TwoDVec m_vecOldBall = null;
	private TwoDVec m_vecBallNow = null;
	/**
	*	@param  nAngle
	*	@return TwoDVec	the predicted position of the ball
	*/
	protected TwoDVec findInterceptBall(int nAngle)
	{
		m_vecOldBall = m_vecBallNow;
		m_vecBallNow = m_Env.getAbsolutePos(m_Env.m_objBall);

		TwoDVec vec = new TwoDVec(m_vecOldBall, m_vecBallNow);
		if(vec.getMagnitude() < 0.1f)
			return m_vecBallNow;

		int nIndex = 0;
		while(++nIndex < 20 && withinRange(Math.abs(angleToPoint(vec.getX(), vec.getY())), 0, nAngle))
			vec.extrapolateNext((float)(m_Env.m_cfgData.ball_decay));

		vec.extrapolateNext(-1.5f);

		if(!withinRange(Math.abs(angleToPoint(vec.getX(), vec.getY())), 0, 120) || vec.getDistance(m_Env.m_vecPosition) < 2)
			return m_vecBallNow;

		return vec;
	}
}
