package betir;

import java.net.SocketException;
import java.io.*;
import java.util.*;
import java.awt.geom.Point2D;

/**
*	This is the class that keeps track of the position of the player
*	and all that he sees, and calculates the absolute position of all objects.
*/
public class RobocupEnvironment
{
//****************************************************************************************************************** member variables

    public static ConfigurationData m_cfgData;
    public int m_nCurrentTime = 0;
    public int m_nZone;

	//robocup soccer playing information
	public ObjectInfo m_objBall;
	public SelfInfo m_objSelf;
	public ArrayList m_aobjFlags;
	public ArrayList m_aobjLines;
	public ArrayList m_aobjFriends;
	public ArrayList m_aobjEnemies;
	public ObjectInfo m_objGoal;
	public ArrayList m_aYells;
	public String m_strPlayMode;
	/** a lock used for certain things when accuarcy is important,
	*	only set to true when new data is recieved from soccer server
	*	and is set to false when a player turns or dashes*/
	public boolean m_bAbsAngle;

	/** max number of yells remembered*/
	private final int m_nYellClear = 10;
	/** the flags*/
	public HashMap m_hashStatic;
	/** position of self*/
	public TwoDVec m_vecPosition = new TwoDVec();
	public boolean m_bHasBall;
	/** a reference to self to allow the adding of behaviors at any time*/
	public Player m_Player;

//****************************************************************************************************************** constructors

    public RobocupEnvironment(ConfigurationData cfgData, boolean display, SelfInfo self) throws SocketException, IOException
    {
		m_cfgData = cfgData;
		m_strPlayMode = "before_kick_off";

		m_aobjFlags = new ArrayList();
		m_aobjLines = new ArrayList();
		m_aobjFriends = new ArrayList();
		m_aobjEnemies = new ArrayList();
		m_aYells = new ArrayList();
        m_objSelf = self;

		initStaticObjects();

		m_bHasBall = false;
    }

//****************************************************************************************************************** parse

	/**
	*	parse the data recieved from the robocup server.
	*
	*	@param data
	*/
	public void parse(String data)
	{
        if (data == null)
        {
            return;
        }
        if (data == "")
        {
            return;
        }
        try
        {
			if(data.charAt(1) == 'p')
            {
				//player type message
			}
            else if(data.charAt(1) == 's')
            {
                if(data.charAt(3) == 'e')
                {
                    //A 'see' message
                    parseSee(data);
                }
                else
                {
                    //A 'sense_body' message
					m_objSelf.parse(data);
                }
            }
            else
            {
                StringTokenizer strtok = new StringTokenizer(data.substring(1, data.length() - 1));
                if(strtok.nextToken().equals("hear"))
                {
                    String timestamp = strtok.nextToken();
                    String sender = strtok.nextToken();
                    if(sender.equals("referee"))
                    {
                        m_strPlayMode = strtok.nextToken(") ");
                    }
                    else
                    {
                        m_aYells.add(strtok.nextToken(") "));
						if(m_aYells.size() > m_nYellClear)
							m_aYells.remove(0);
                    }
                }
                else
                {
                    System.out.println("ERROR: Unknown message: " + data);
                }
            }
		}
        catch (StringIndexOutOfBoundsException e)
        {
            System.out.println("ERROR: RobocupEnvironment.parse(). Could not parse:\n" + data);
            return;
        }
    }

	private void parseSee(String data)
	{
		// init the string tokenizer with the data sans the outer '(' & ')'
        StringTokenizer strtok = new StringTokenizer(data.substring(1, data.length()-1), "(");
        String token = strtok.nextToken();

        // verify that this is indeed a 'see' message
        if(token.startsWith("see"))
        {
            m_aobjFriends.clear();
			m_aobjEnemies.clear();
			m_aobjFlags.clear();
			m_aobjLines.clear();
			m_objGoal = null;
			m_bHasBall = false;
			m_objBall = null;

			try
			{
				m_nCurrentTime = Integer.parseInt(token.substring(4).trim());
			}
			catch(NumberFormatException e)
			{
				// uh oh, probably out of bounds
			};

			String objName;
			String objInfo;
			StringTokenizer strtokObjInfo;
			PlayerInfo theObject;

            // parse the seen objects
            while(strtok.hasMoreTokens())
            {
				theObject = null;

                // the next token is the object name
                objName = strtok.nextToken(")");

                // if there are preceeding '(' replace with spaces and then trim
                objName = objName.replace('(',' ').trim();
                theObject = new PlayerInfo(objName, 0, m_nCurrentTime);
                try
                {
                    objInfo = strtok.nextToken();
                }
                catch (NoSuchElementException e)
                {
                    break;
                }
				strtokObjInfo = new StringTokenizer(objInfo);
                for(int i = 1; strtokObjInfo.countTokens() > 0; ++i)
                {
                    // the cases below indicate position number; for instance, distance
                    // is always the first argument, direction the second,...
                    // of course not every argument will appear for each object... the outer
                    // loop takes care of this issue.
                    float num;
                    try
                    {
                        num = Float.parseFloat(strtokObjInfo.nextToken());
                    }
                    catch (NumberFormatException e)
                    {
                        //its "nan", so set it to 0 instead
                        num = 0;
                    }

                    switch( i )
                    {
                    case 1:
                        theObject.setDistance(num);
                        break;
                    case 2:
                        theObject.setAngle((int)num);
                        break;
                    case 3:
                        theObject.setDistChng(num);
                        theObject.setGotChanges(true);
                        break;
                    case 4:
                        theObject.setAngleChng((int)num);
                        break;
                    case 5:
                        theObject.setBodyAngle((int)num);
                        break;
                    case 6:
                        theObject.setHeadAngle((int)num);
                        break;
                    }
                }

				switch(objName.charAt(0))
				{
				case 'b':
				case 'B':
					m_objBall = (ObjectInfo)theObject;
					if(m_objBall.getDistance() < m_cfgData.kickable_area * 0.75)
						m_bHasBall = true;
					//System.err.println("see ball: " + data);
					break;
				case 'p':
				case 'P':
					if(objName.length() > 2)
					{
						if(objName.charAt(objName.length() - 1) == '"')
							objName = objName.substring(3, objName.length() - 1);
						else
							objName = objName.substring(3, objName.length() - 3);
						if(objName.charAt(objName.length() - 1) == '"')
							objName = objName.substring(0, objName.length() - 1);
						if(objName.equals(m_objSelf.getTeam()))
							m_aobjFriends.add(theObject);
						else
							m_aobjEnemies.add(theObject);
					}
					else
						m_aobjEnemies.add(theObject);
					//System.err.println("see player \"" + objName + "\"");
					break;
				case 'f':
				case 'F':
					if(objName.length() > 2)
					{
						m_aobjFlags.add(theObject);
						if(objName.equals("f l 0") && m_objSelf.getSide() == 'r')
							m_objGoal = (ObjectInfo)theObject;
						else if(objName.equals("f r 0") && m_objSelf.getSide() == 'l')
							m_objGoal = (ObjectInfo)theObject;
					}
					//System.err.println("see flag \"" + objName + "\"");
					break;
				case 'g':
				case 'G':
					//if(objName.length() > 2 && objName.charAt(2) != m_objSelf.getSide())
					//	m_objGoal = (ObjectInfo)theObject;
					//System.err.println("see goal \"" + objName + "\"");
					break;
				case 'l':
				case 'L':
					m_aobjLines.add(theObject);
					//System.err.println("see line \"" + objName + "\"");
					break;
				default:
					//System.err.println("ERROR: do not know this object: \"" + objName + "\"");
					break;
				}
            }
        }
        else
            System.err.println("ERROR: ParseSee: String does not have see: " + data);

		try
		{
        	CalcAbsolute();
		}
		catch(Exception e)
		{
			System.err.println("Error in CalcAbsolute");
			e.printStackTrace();
			debug();
		}
	}


//****************************************************************************************************************** debug

	/**
	*	prints out info useful for debuging
	*/
	public void debug()
	{
		System.out.println("-----------------------");
		if(m_objBall != null)
			System.out.println("ball");
		if(m_objGoal != null)
			System.out.println("goal");
		System.out.println(m_strPlayMode);
		System.out.print("flags: ");
		for(int i = 0; i < m_aobjFlags.size(); ++i)
			System.out.print(((ObjectInfo)m_aobjFlags.get(i)).getName() + ", ");
		System.out.println(" ");
		System.out.print("lines: ");
		for(int i = 0; i < m_aobjLines.size(); ++i)
			System.out.print(((ObjectInfo)m_aobjLines.get(i)).getName() + ", ");
		System.out.println(" ");
		System.out.print("friends: ");
		for(int i = 0; i < m_aobjFriends.size(); ++i)
			System.out.print(((ObjectInfo)m_aobjFriends.get(i)).getName() + ", ");
		System.out.println(" ");
		System.out.print("enemies: ");
		for(int i = 0; i < m_aobjEnemies.size(); ++i)
			System.out.print(((ObjectInfo)m_aobjEnemies.get(i)).getName() + ", ");
		System.out.println(" ");
		System.out.print("yells: ");
		for(int i = 0; i < m_aYells.size(); ++i)
			System.out.print(((ObjectInfo)m_aYells.get(i)).getName() + ", ");
		System.out.println(" ");
		System.out.println("absolute pos: " + m_vecPosition.getX() + ", " + m_vecPosition.getY());
		System.out.println("angle: " + m_vecPosition.getAngle());
	}

//****************************************************************************************************************** calc absolute

	private ObjectInfo m_flagF1;
	private ObjectInfo m_flagF2;
	/*
	*	Calculates the absolute postion of the player.
	*/
	public void CalcAbsolute()
	{
		m_bAbsAngle = true;
		if(m_aobjFlags.size() > 1)
		{
			//find angle facing
			m_flagF1 = (ObjectInfo)(m_aobjFlags.get(0));
			m_flagF2 = (ObjectInfo)(m_aobjFlags.get(1));

			for(int i = 0; i < m_aobjFlags.size(); ++i)
			{
				if(((ObjectInfo)m_aobjFlags.get(i)).getDistance() < m_flagF1.getDistance())
				{
					m_flagF2 = m_flagF1;
					m_flagF1 = (ObjectInfo)(m_aobjFlags.get(i));
				}
			}

			float fXChange = (float)(((Point2D.Float)(m_hashStatic.get(m_flagF2.getName()))).getX() - ((Point2D.Float)(m_hashStatic.get(m_flagF1.getName()))).getX());
			float fYChange = (float)(((Point2D.Float)(m_hashStatic.get(m_flagF2.getName()))).getY() - ((Point2D.Float)(m_hashStatic.get(m_flagF1.getName()))).getY());
			int nD = QuickIntMath.Sqrt((int)(fXChange * fXChange + fYChange * fYChange));
			int nOffset = 0;
			if (nD >= m_flagF2.getDistance() + m_flagF1.getDistance())
				nOffset = 0;
			else if (m_flagF1.getDistance() >= nD + m_flagF2.getDistance())
				nOffset = 180;
			else if (m_flagF2.getDistance() >= nD + m_flagF1.getDistance())
				nOffset = 0;
			else
				nOffset = (int)(57.29577 * Math.acos((nD * nD + m_flagF2.getDistance() * m_flagF2.getDistance() - m_flagF1.getDistance() * m_flagF1.getDistance()) / (2.0f * m_flagF2.getDistance() * nD)));

			int nAngle = 0;
			if (fXChange == 0)
				nAngle = (fYChange >= 0) ? 90 : 270;
			else
			{
				nAngle = (int)(57.29577 * Math.atan(fYChange / fXChange));
				if (fXChange < 0)
					nAngle += 180;
			}

			if(m_flagF1.getAngle() < m_flagF2.getAngle())
			{
				m_vecPosition.setAngle(nAngle + nOffset + m_flagF2.getAngle());
			}
			else
			{
				m_vecPosition.setAngle(nAngle - nOffset + m_flagF2.getAngle());
			}

			//find position
			float fYOffset = QuickIntMath.Sin(m_vecPosition.getAngle() - m_flagF1.getAngle()) * m_flagF1.getDistance();
			float fXOffset = QuickIntMath.Cos(m_vecPosition.getAngle() - m_flagF1.getAngle()) * m_flagF1.getDistance();

			Point2D.Float objLocation = (Point2D.Float)m_hashStatic.get(m_flagF1.getName());

			if (objLocation != null)
			{
				m_vecPosition.setAbsolute((int)(objLocation.getX() - fXOffset), (int)(objLocation.getY() - fYOffset));
			}
		}
	}

//****************************************************************************************************************** get pos

	/*
	*	uses the absolute position of the player to get the absolute position of anything he sees.
	*/
	public TwoDVec getAbsolutePos(ObjectInfo theObj)
	{

		int nY = (int)(QuickIntMath.Sin(m_vecPosition.getAngle() - theObj.getAngle()) * theObj.getDistance());
		int nX = (int)(QuickIntMath.Cos(m_vecPosition.getAngle() - theObj.getAngle()) * theObj.getDistance());
		TwoDVec vec = new TwoDVec(m_vecPosition.getX() + nX, m_vecPosition.getY() + nY, 0, 0);
		if(m_objSelf.getSide() == 'r')
			vec.setAbsolute(-vec.getX(), vec.getY());
		return vec;
	}

//****************************************************************************************************************** initstatic

	/**
    * Sets the location of all the flags.
    */
    private void initStaticObjects()
    {
        m_hashStatic = new HashMap();

        // perimeter flags clockwise from lower-left
        m_hashStatic.put("f l b", new Point2D.Float(-52.5f, -34f));
        m_hashStatic.put("f l b 30", new Point2D.Float(-57.5f, -30f));
        m_hashStatic.put("f l b 20", new Point2D.Float(-57.5f, -20f));
        m_hashStatic.put("f l b 10", new Point2D.Float(-57.5f, -10f));
        m_hashStatic.put("f l 0", new Point2D.Float(-57.5f, 0f));
        m_hashStatic.put("f l t 10", new Point2D.Float(-57.5f, 10f));
        m_hashStatic.put("f l t 20", new Point2D.Float(-57.5f, 20f));
        m_hashStatic.put("f l t 30", new Point2D.Float(-57.5f, 30f));
        m_hashStatic.put("f l t", new Point2D.Float(-52.5f, 34f));
        m_hashStatic.put("f t l 50", new Point2D.Float(-50f, 39f));
        m_hashStatic.put("f t l 40", new Point2D.Float(-40f, 39f));
        m_hashStatic.put("f t l 30", new Point2D.Float(-30f, 39f));
        m_hashStatic.put("f t l 20", new Point2D.Float(-20f, 39f));
        m_hashStatic.put("f t l 10", new Point2D.Float(-10f, 39f));
        m_hashStatic.put("f t 0", new Point2D.Float(0f, 39f));
        m_hashStatic.put("f t r 10", new Point2D.Float(10f, 39f));
        m_hashStatic.put("f t r 20", new Point2D.Float(20f, 39f));
        m_hashStatic.put("f t r 30", new Point2D.Float(30f, 39f));
        m_hashStatic.put("f t r 40", new Point2D.Float(40f, 39f));
        m_hashStatic.put("f t r 50", new Point2D.Float(50f, 39f));
        m_hashStatic.put("f r t", new Point2D.Float(52.5f, 34f));
        m_hashStatic.put("f r t 30", new Point2D.Float(57.5f, 30f));
        m_hashStatic.put("f r t 20", new Point2D.Float(57.5f, 20f));
        m_hashStatic.put("f r t 10", new Point2D.Float(57.5f, 10f));
        m_hashStatic.put("f r 0", new Point2D.Float(57.5f, 0f));
        m_hashStatic.put("f r b 10", new Point2D.Float(57.5f, -10f));
        m_hashStatic.put("f r b 20", new Point2D.Float(57.5f, -20f));
        m_hashStatic.put("f r b 30", new Point2D.Float(57.5f, -30f));
        m_hashStatic.put("f r b", new Point2D.Float(52.5f, -34f));
        m_hashStatic.put("f b r 50", new Point2D.Float(50f, -39f));
        m_hashStatic.put("f b r 40", new Point2D.Float(40f, -39f));
        m_hashStatic.put("f b r 30", new Point2D.Float(30f, -39f));
        m_hashStatic.put("f b r 20", new Point2D.Float(20f, -39f));
        m_hashStatic.put("f b r 10", new Point2D.Float(10f, -39f));
        m_hashStatic.put("f b 0", new Point2D.Float(0f, -39f));
        m_hashStatic.put("f b l 10", new Point2D.Float(-10f, -39f));
        m_hashStatic.put("f b l 20", new Point2D.Float(-20f, -39f));
        m_hashStatic.put("f b l 30", new Point2D.Float(-30f, -39f));
        m_hashStatic.put("f b l 40", new Point2D.Float(-40f, -39f));
        m_hashStatic.put("f b l 50", new Point2D.Float(-50f, -39f));
        // goals and penalty area flags
        // note: approximated the width of the penalty box to be 16 yds
        m_hashStatic.put("g l", new Point2D.Float(-52.5f, 0f));
        m_hashStatic.put("f g l b", new Point2D.Float(-52.5f, (float)(-m_cfgData.goal_width/2f)));
        m_hashStatic.put("f g l t", new Point2D.Float(-52.5f, (float)(m_cfgData.goal_width/2f)));
        m_hashStatic.put("f p l b", new Point2D.Float(-36.5f, -20f));
        m_hashStatic.put("f p l c", new Point2D.Float(-36.5f, 0f));
        m_hashStatic.put("f p l t", new Point2D.Float(-36.5f, 20f));
        m_hashStatic.put("g r", new Point2D.Float(52.5f, 0f));
        m_hashStatic.put("f g r b", new Point2D.Float(52.5f, (float)(-m_cfgData.goal_width/2f)));
        m_hashStatic.put("f g r t", new Point2D.Float(52.5f, (float)(m_cfgData.goal_width/2f)));
        m_hashStatic.put("f p r b", new Point2D.Float(36.5f, -20f));
        m_hashStatic.put("f p r c", new Point2D.Float(36.5f, 0f));
        m_hashStatic.put("f p r t", new Point2D.Float(36.5f, 20f));
        // center line flags
        m_hashStatic.put("f c", new Point2D.Float(0f, 0f));
        m_hashStatic.put("f c b", new Point2D.Float(0f, -34f));
        m_hashStatic.put("f c t", new Point2D.Float(0f, 34f));
    }
}
